package org.cometd.client;

import java.util.LinkedList;
import java.util.Map;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeUnit;
import javax.servlet.http.Cookie;

import junit.framework.TestCase;
import org.cometd.Bayeux;
import org.cometd.Client;
import org.cometd.Message;
import org.cometd.MessageListener;
import org.cometd.server.AbstractBayeux;
import org.cometd.server.MessageImpl;
import org.cometd.server.continuation.ContinuationCometdServlet;
import org.eclipse.jetty.client.HttpClient;
import org.eclipse.jetty.server.Server;
import org.eclipse.jetty.server.handler.HandlerCollection;
import org.eclipse.jetty.server.nio.SelectChannelConnector;
import org.eclipse.jetty.servlet.ServletContextHandler;
import org.eclipse.jetty.servlet.ServletHolder;

/**
 * Simulates a browser opening multiple tabs to the same Bayeux server
 */
public class MultipleClientSessionsTest extends TestCase
{
    private Server server;
    private String cometdURL;
    private long timeout = 5000L;
    private AbstractBayeux bayeux;
    private HttpClient client;

    @Override
    protected void setUp() throws Exception
    {
        server = new Server();
        SelectChannelConnector connector = new SelectChannelConnector();
        server.addConnector(connector);

        HandlerCollection handlers = new HandlerCollection();
        server.setHandler(handlers);

        String contextPath = "/cometd";
        ServletContextHandler context = new ServletContextHandler(handlers, contextPath, ServletContextHandler.SESSIONS);

        // Setup comet servlet
        ContinuationCometdServlet cometdServlet = new ContinuationCometdServlet();
        ServletHolder cometdServletHolder = new ServletHolder(cometdServlet);
        cometdServletHolder.setInitParameter("timeout", String.valueOf(timeout));
        cometdServletHolder.setInitParameter("logLevel", "3");
        String cometdServletPath = "/cometd";
        context.addServlet(cometdServletHolder, cometdServletPath + "/*");

        server.start();

        int port = connector.getLocalPort();
        String contextURL = "http://localhost:" + port + contextPath;
        cometdURL = contextURL + cometdServletPath;

        bayeux = cometdServlet.getBayeux();

        client = new HttpClient();
        client.start();
    }

    @Override
    protected void tearDown() throws Exception
    {
        client.stop();

        server.stop();
        server.join();
    }

    public void testMultipleClientSession_WithNoMultiSessionInterval() throws Exception
    {
        bayeux.setMultiFrameInterval(-1);

        final CountDownLatch connectLatch1 = new CountDownLatch(1);
        BayeuxClient client1 = new BayeuxClient(client, cometdURL)
        {
            @Override
            protected void metaConnect(boolean success, Message message)
            {
                if (success)
                    connectLatch1.countDown();
            }
        };
        final ConcurrentLinkedQueue<Message> connects1 = new ConcurrentLinkedQueue<Message>();
        final CountDownLatch latch1 = new CountDownLatch(2);
        client1.addListener(new MessageListener()
        {
            public void deliver(Client from, Client to, Message message)
            {
                if (Bayeux.META_CONNECT.equals(message.getChannel()))
                {
                    connects1.offer(message);
                    ((MessageImpl)message).incRef();
                    latch1.countDown();
                }
            }
        });
        client1.start();
        try
        {
            assertTrue(connectLatch1.await(1000, TimeUnit.MILLISECONDS));
            Cookie cookie = client1.getCookie("BAYEUX_BROWSER");
            assertNotNull(cookie);

            // Give some time to the first client to establish the long poll before the second client
            Thread.sleep(500);

            BayeuxClient client2 = new BayeuxClient(client, cometdURL){};
            final ConcurrentLinkedQueue<Message> connects2 = new ConcurrentLinkedQueue<Message>();
            final CountDownLatch latch2 = new CountDownLatch(2);
            client2.setCookie(cookie);
            client2.addListener(new MessageListener()
            {
                public void deliver(Client from, Client to, Message message)
                {
                    if (Bayeux.META_CONNECT.equals(message.getChannel()))
                    {
                        connects2.offer(message);
                        ((MessageImpl)message).incRef();
                        latch2.countDown();
                    }
                }
            });
            client2.start();

            assertTrue(latch2.await(1000, TimeUnit.MILLISECONDS));
            assertEquals(2, connects2.size());
            connects2.remove();
            Message secondConnect2 = connects2.remove();
            Advice advice = new Advice((Map)secondConnect2.get("advice"));
            assertTrue(advice.isReconnectNone());
            assertTrue(advice.isMultipleClients());

            // Give some time to the second client to process the disconnect
            Thread.sleep(500);
            assertFalse(client2.isRunning());

            assertTrue(latch1.await(timeout, TimeUnit.MILLISECONDS));
            assertEquals(2, connects1.size());
            assertTrue(client1.isRunning());
        }
        finally
        {
            client1.stop();
        }
    }

    public void testMultipleClientSession_WithMultiSessionInterval() throws Exception
    {
        Long multiSessionInterval = timeout / 5;
        bayeux.setMultiFrameInterval(multiSessionInterval.intValue());

        final CountDownLatch connectLatch1 = new CountDownLatch(1);
        BayeuxClient client1 = new BayeuxClient(client, cometdURL)
        {
            @Override
            protected void metaConnect(boolean success, Message message)
            {
                System.out.println("1 - message = " + message);
                if (success)
                    connectLatch1.countDown();
            }
        };
        final ConcurrentLinkedQueue<Message> connects1 = new ConcurrentLinkedQueue<Message>();
        client1.addListener(new MessageListener()
        {
            public void deliver(Client client, Client client1, Message message)
            {
                if (Bayeux.META_CONNECT.equals(message.getChannel()))
                {
                    connects1.offer(message);
                    ((MessageImpl)message).incRef();
                }
            }
        });
        client1.start();
        assertTrue(connectLatch1.await(1000, TimeUnit.MILLISECONDS));
        Cookie cookie = client1.getCookie("BAYEUX_BROWSER");
        assertNotNull(cookie);

        // Give some time to the first client to establish the long poll before the second client
        Thread.sleep(500);

        System.out.println("client1 done");

        final CountDownLatch connectLatch2 = new CountDownLatch(1);
        BayeuxClient client2 = new BayeuxClient(client, cometdURL)
        {
            @Override
            protected void metaConnect(boolean success, Message message)
            {
                System.out.println("2 - message = " + message);
                if (success)
                    connectLatch2.countDown();
            }
        };
        final ConcurrentLinkedQueue<Message> connects2 = new ConcurrentLinkedQueue<Message>();
        client2.setCookie(cookie);
        client2.addListener(new MessageListener()
        {
            public void deliver(Client client, Client client1, Message message)
            {
                if (Bayeux.META_CONNECT.equals(message.getChannel()))
                {
                    connects2.offer(message);
                    ((MessageImpl)message).incRef();
                }
            }
        });
        client2.start();
        assertTrue(connectLatch2.await(1000, TimeUnit.MILLISECONDS));

        Thread.sleep(500);

        System.out.println("client2 done");

        final CountDownLatch connectLatch3 = new CountDownLatch(1);
        BayeuxClient client3 = new BayeuxClient(client, cometdURL)
        {
            @Override
            protected void metaConnect(boolean success, Message message)
            {
                System.out.println("3 - message = " + message);
                if (success)
                    connectLatch3.countDown();
            }
        };
        final ConcurrentLinkedQueue<Message> connects3 = new ConcurrentLinkedQueue<Message>();
        client3.setCookie(cookie);
        client3.addListener(new MessageListener()
        {
            public void deliver(Client client, Client client1, Message message)
            {
                if (Bayeux.META_CONNECT.equals(message.getChannel()))
                {
                    connects3.offer(message);
                    ((MessageImpl)message).incRef();
                }
            }
        });
        client3.start();
        assertTrue(connectLatch3.await(1000, TimeUnit.MILLISECONDS));

        // Sleep for a while
        Thread.sleep(2 * multiSessionInterval);

        // The first client must remain in long poll mode
        assertEquals(1, connects1.size());

        // Second client must be in normal poll mode
        assertTrue(connects2.size() > 1);
        Message lastConnect2 = new LinkedList<Message>(connects2).getLast();
        Map advice2Map = (Map)lastConnect2.get(Bayeux.ADVICE_FIELD);
        assertNotNull(advice2Map);
        Advice advice2 = new Advice(advice2Map);
        assertTrue(advice2.isMultipleClients());

        // Third client must be in normal poll mode
        assertTrue(connects3.size() > 1);
        Message lastConnect3 = new LinkedList<Message>(connects3).getLast();
        Map advice3Map = (Map)lastConnect3.get(Bayeux.ADVICE_FIELD);
        assertNotNull(advice3Map);
        Advice advice3 = new Advice(advice3Map);
        assertTrue(advice3.isMultipleClients());

        // Wait for the first client to re-issue a long poll
        Thread.sleep(timeout);

        // First client must still be in long poll mode
        assertEquals(2, connects1.size());

        // Abort abruptly the first client
        // Another client must switch to long poll
        client1.abort();

        // Sleep until the first client times out and disappears from server
        Thread.sleep(bayeux.getMaxInterval() + timeout);

        // Be sure client1 did not poll
        assertEquals(2, connects1.size());

        // Loop until one of the other clients switched to long poll
        BayeuxClient client4 = null;
        BayeuxClient client5 = null;
        for (int i = 0; i < 10; ++i)
        {
            lastConnect2 = new LinkedList<Message>(connects2).getLast();
            advice2Map = (Map)lastConnect2.get(Bayeux.ADVICE_FIELD);
            if (advice2Map == null || !advice2Map.containsKey("multiple-clients"))
            {
                client4 = client2;
                client5 = client3;
                break;
            }
            lastConnect3 = new LinkedList<Message>(connects3).getLast();
            advice3Map = (Map)lastConnect3.get(Bayeux.ADVICE_FIELD);
            if (advice3Map == null || !advice3Map.containsKey("multiple-clients"))
            {
                client4 = client3;
                client5 = client2;
                break;
            }
            Thread.sleep(timeout / 10);
        }
        assertNotNull(client4);

        // Disconnect this client normally, the last client must switch to long poll
        client4.disconnect();
        Thread.sleep(500);

        // Be sure the last client had the time to switch to long poll mode
        Thread.sleep(timeout + multiSessionInterval);
        Message lastConnect;
        if (client5 == client2)
            lastConnect = new LinkedList<Message>(connects2).getLast();
        else
            lastConnect = new LinkedList<Message>(connects3).getLast();
        Map advice = (Map)lastConnect.get(Bayeux.ADVICE_FIELD);
        assertTrue(advice == null || !advice.containsKey("multiple-clients"));

        client5.disconnect();
        Thread.sleep(500);
    }
}
