// ========================================================================
// Copyright 2004-2008 Mort Bay Consulting Pty. Ltd.
// ------------------------------------------------------------------------
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
// http://www.apache.org/licenses/LICENSE-2.0
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// ========================================================================

package org.cometd.javascript;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.net.URL;

import junit.framework.TestCase;
import org.cometd.server.AbstractBayeux;
import org.cometd.server.continuation.ContinuationCometdServlet;
import org.eclipse.jetty.server.Server;
import org.eclipse.jetty.server.handler.HandlerCollection;
import org.eclipse.jetty.server.nio.SelectChannelConnector;
import org.eclipse.jetty.servlet.DefaultServlet;
import org.eclipse.jetty.servlet.ServletContextHandler;
import org.eclipse.jetty.servlet.ServletHolder;
import org.mozilla.javascript.ScriptableObject;

/**
 * @version $Revision: 1483 $ $Date: 2009-03-04 14:56:47 +0100 (Wed, 04 Mar 2009) $
 */
public abstract class AbstractCometdTest extends TestCase
{
    private HttpCookieStore cookies;
    private Server server;
    protected int longPollingPeriod = 5000;
    protected String cometServletPath = "/cometd";
    protected int port;
    protected String contextURL;
    protected String cometdURL;
    protected int expirationPeriod = 2500;
    private ThreadModel threadModel;
    private XMLHttpRequestClient xhrClient;

    @Override
    protected void setUp() throws Exception
    {
        cookies = new HttpCookieStore();
        initCometServer();
    }

    public void initCometServer() throws Exception
    {
        server = new Server();
        SelectChannelConnector connector = new SelectChannelConnector();
        server.addConnector(connector);

        HandlerCollection handlers = new HandlerCollection();
        server.setHandler(handlers);

        String contextPath = "/cometd";
        ServletContextHandler context = new ServletContextHandler(handlers, contextPath, ServletContextHandler.SESSIONS);

        // Setup default servlet to serve static files
        context.addServlet(DefaultServlet.class, "/");

        // Setup comet servlet
        ContinuationCometdServlet cometServlet = new ContinuationCometdServlet();
        ServletHolder cometServletHolder = new ServletHolder(cometServlet);
        cometServletHolder.setInitParameter("timeout", String.valueOf(longPollingPeriod));
        cometServletHolder.setInitParameter("logLevel", "2");
        cometServletHolder.setInitParameter("requestAvailable", "true");
        cometServletHolder.setInitParameter("maxInterval", String.valueOf(expirationPeriod));
        context.addServlet(cometServletHolder, cometServletPath + "/*");

        customizeContext(context);

        server.start();
        port = connector.getLocalPort();

        contextURL = "http://localhost:" + port + contextPath;
        cometdURL = contextURL + cometServletPath;

        customizeBayeux(cometServlet.getBayeux());
    }

    public void initJavaScript() throws Exception
    {
        // Initializes the thread model
        org.mozilla.javascript.Context jsContext = org.mozilla.javascript.Context.enter();
        try
        {
            ScriptableObject rootScope = jsContext.initStandardObjects();

            ScriptableObject.defineClass(rootScope, HttpCookieStore.class);
            jsContext.evaluateString(rootScope, "var cookies = new HttpCookieStore();", "cookies", 1, null);
            HttpCookieStore cookies = (HttpCookieStore)rootScope.get("cookies", rootScope);
            cookies.putAll(this.cookies);
            this.cookies = cookies;

            ScriptableObject.defineClass(rootScope, JavaScriptThreadModel.class);
            jsContext.evaluateString(rootScope, "var threadModel = new JavaScriptThreadModel(this);", "threadModel", 1, null);
            threadModel = (ThreadModel)rootScope.get("threadModel", rootScope);
            threadModel.init();

            ScriptableObject.defineClass(rootScope, XMLHttpRequestClient.class);
            ScriptableObject.defineClass(rootScope, XMLHttpRequestExchange.class);
            jsContext.evaluateString(rootScope, "var xhrClient = new XMLHttpRequestClient(2);", "xhrClient", 1, null);
            xhrClient = (XMLHttpRequestClient)rootScope.get("xhrClient", rootScope);
            xhrClient.start();
        }
        finally
        {
            org.mozilla.javascript.Context.exit();
        }
    }

    @Override
    protected void tearDown() throws Exception
    {
        destroyCometServer();
        cookies.clear();
    }

    public void destroyJavaScript() throws Exception
    {
        threadModel.destroy();
        xhrClient.stop();
    }

    public void destroyCometServer() throws Exception
    {
        server.stop();
        server.join();
    }

    protected abstract void customizeContext(ServletContextHandler context) throws Exception;

    protected void customizeBayeux(AbstractBayeux bayeux)
    {
    }

    protected void evaluateURL(URL script) throws IOException
    {
        threadModel.evaluate(script);
    }

    @SuppressWarnings("unchecked")
    protected <T> T evaluateScript(String script)
    {
        return (T)evaluateScript(null, script);
    }

    @SuppressWarnings("unchecked")
    protected <T> T evaluateScript(String scriptName, String script)
    {
        return (T)threadModel.evaluate(scriptName, script);
    }

    protected void defineClass(Class clazz) throws InvocationTargetException, InstantiationException, IllegalAccessException
    {
        threadModel.define(clazz);
    }

    @SuppressWarnings("unchecked")
    protected <T> T get(String name)
    {
        return (T) threadModel.get(name);
    }
}
